`timescale 1ns / 1ps
`default_nettype none

//******************************************************************************
//* Shiftregiszter alap FIFO bellthat mlysggel s sz mrettel.          *
//******************************************************************************
module fifo #(
   //A FIFO mlysge.
   parameter DEPTH = 16,
   //A szavak mrete bitekben.
   parameter WIDTH = 8
) (
   //rajel s reset.
   input wire              clk,        //rajel.
   input wire              rst,        //Reset jel.
   
   //Adat be- s kimenetek.
   input  wire [WIDTH-1:0] data_in,    //FIFO adat bemenet.
   output wire [WIDTH-1:0] data_out,   //FIFO adat kimenet.
   
   //Vezrl bemenetek.
   input  wire             write,      //FIFO rs engedlyez jel.
   input  wire             read,       //FIFO olvass engedlyez jel.
   
   //llapotjelz kimenetek.
   output reg              exists,     //A FIFO nem res.
   output reg              full        //A FIFO tele van.
);

//******************************************************************************
//* Kettes alap logaritmus fggvny.                                          *
//******************************************************************************
function integer log2(input integer x);
begin
   x = x - 1;
   for (log2 = 0; x > 0; log2 = log2 + 1)
      x = x >> 1;
end
endfunction


//******************************************************************************
//* Az olvassi cmszmll bitjeinek szma.                                   *
//******************************************************************************
localparam CNT_LEN = log2(DEPTH);


//******************************************************************************
//* Dinamikus shiftregiszter.                                                  *
//******************************************************************************
reg  [DEPTH-1:0]   fifo_shr [WIDTH-1:0];
reg  [CNT_LEN-1:0] rd_addr;
wire               shr_en = write & (~full | read);

genvar i;

generate
   for (i = 0; i < WIDTH; i = i + 1)
   begin: shr_loop
      always @(posedge clk)
      begin
         if (shr_en)
            fifo_shr[i] <= {fifo_shr[i][DEPTH-2:0], data_in[i]};
      end
      
      assign data_out[i] = fifo_shr[i][rd_addr];
   end
endgenerate


//******************************************************************************
//* Olvassi cmszmll.                                                      *
//******************************************************************************
wire rd_addr_min = (rd_addr == 0);
wire rd_addr_max = (rd_addr == (DEPTH - 1));
wire rd_addr_inc = ~rd_addr_max &  write & ~read & exists;
wire rd_addr_dec = ~rd_addr_min & ~write &  read;

always @(posedge clk)
begin
   if (rst)
      rd_addr <= 0;
   else
      if (rd_addr_inc)
         rd_addr <= rd_addr + 1;
      else
         if (rd_addr_dec)
            rd_addr <= rd_addr - 1;
end


//******************************************************************************
//* FIFO llapotjelz bitek.                                                   *
//******************************************************************************
//FIFO nem res jelzs.
wire exists_clr = rd_addr_min & read & ~write; 
wire exists_set = write;

always @(posedge clk)
begin
   if (rst || exists_clr)
      exists <= 1'b0;
   else
      if (exists_set)
         exists <= 1'b1;
end

//FIFO tele jelzs.
wire full_set = (rd_addr == (DEPTH - 2)) & write & ~read;
wire full_clr = ~write & read;

always @(posedge clk)
begin
   if (rst || full_clr)
      full <= 1'b0;
   else
      if (full_set)
         full <= 1'b1;
end

endmodule

`default_nettype wire
